@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: axis.c,v 1.223.2.4 2017/09/03 03:38:36 sfeam Exp $"); }
+static char *RCSid() { return RCSid("$Id: axis.c,v 1.223.2.13 2017/10/12 05:01:43 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - axis.c */
@@ -208,12 +208,9 @@ axis_unlog_interval(struct axis *axis, double *min, double *max, TBOOLEAN checkr
 void
 axis_invert_if_requested(struct axis *axis)
 {
-    if (((axis->range_flags & RANGE_IS_REVERSED))
-    &&  (axis->autoscale != 0) &&  (axis->max > axis->min) ) {
-	double temp = axis->min;
-	axis->min = axis->max;
-	axis->max = temp;
-    }
+    if (((axis->range_flags & RANGE_IS_REVERSED)) &&  (axis->autoscale != 0))
+	/* NB: The whole point of this is that we want max < min !!! */
+	reorder_if_necessary(axis->max, axis->min);
 }
 
 
@@ -1125,16 +1122,12 @@ gen_tics(struct axis *this, tic_callback callback)
 	double user;		/* in user co-ords */
 	double start, step, end;
 	int    nsteps;
-	double lmin = this->min, lmax = this->max;
 	double internal_min, internal_max;	/* to allow for rounding errors */
 	double ministart = 0, ministep = 1, miniend = 1;	/* internal or user - depends on step */
+	double lmin = this->min, lmax = this->max;
+
+	reorder_if_necessary(lmin, lmax);
 
-	if (lmax < lmin) {
-	    /* hmm - they have set reversed range for some reason */
-	    double temp = lmin;
-	    lmin = lmax;
-	    lmax = temp;
-	}
 	/* {{{  choose start, step and end */
 	switch (def->type) {
 	case TIC_SERIES:
@@ -1151,6 +1144,22 @@ gen_tics(struct axis *this, tic_callback callback)
 		    end = step * ceil(lmax / step);
 		else
 		    end = axis_do_log(this, def->def.series.end);
+		if (def->logscaling) {
+		    /* This tries to emulate earlier gnuplot versions in handling
+		     *     set log y; set ytics 10
+		     */
+		    if (start <= 0) {
+			start = step;
+			while (start > this->linked_to_primary->min)
+			    start -= step;
+		    } else {
+			start = eval_link_function(this->linked_to_primary, start);
+		    }
+		    step  = eval_link_function(this->linked_to_primary, step);
+		    end   = eval_link_function(this->linked_to_primary, end);
+		    lmin = this->linked_to_primary->min;
+		    lmax = this->linked_to_primary->max;
+		}
 	    } else {
 		start = def->def.series.start;
 		step = def->def.series.incr;
@@ -1165,6 +1174,7 @@ gen_tics(struct axis *this, tic_callback callback)
 	    if (nonlinear(this)) {
 		lmin = this->linked_to_primary->min;
 		lmax = this->linked_to_primary->max;
+		reorder_if_necessary(lmin, lmax);
 		this->ticstep = make_tics(this->linked_to_primary, 20);
 		/* It may be that we _always_ want ticstep = 1.0 */
 		if (this->ticstep < 1.0)
@@ -1195,15 +1205,8 @@ gen_tics(struct axis *this, tic_callback callback)
 	}
 	/* }}} */
 
-	/* {{{  ensure ascending order */
-	if (end < start) {
-	    double temp;
-	    temp = end;
-	    end = start;
-	    start = temp;
-	}
+	reorder_if_necessary(start, end);
 	step = fabs(step);
-	/* }}} */
 
 	if ((minitics != MINI_OFF) && (this->miniticscale != 0)) {
 	    FPRINTF((stderr,"axis.c: %d  start = %g end = %g step = %g base = %g\n", 
@@ -1339,7 +1342,9 @@ gen_tics(struct axis *this, tic_callback callback)
 		internal = tic;
 		user = tic;
 	    } else if (nonlinear(this)) {
-		if (def->type == TIC_COMPUTED)
+		if (def->type == TIC_SERIES && def->logscaling)
+		    user = eval_link_function(this->linked_to_primary->linked_to_secondary, tic);
+		else if (def->type == TIC_COMPUTED)
 		    user = eval_link_function(this->linked_to_primary->linked_to_secondary, tic);
 		else
 		    user = tic;
@@ -1409,7 +1414,9 @@ gen_tics(struct axis *this, tic_callback callback)
 			}
 
 			/* This is where we finally decided to put the tic mark */
-			if (nonlinear(this) && (def->type == TIC_COMPUTED))
+			if (nonlinear(this) && (def->type == TIC_SERIES && def->logscaling))
+			    position = user;
+			else if (nonlinear(this) && (def->type == TIC_COMPUTED))
 			    position = user;
 			else
 			    position = internal;
@@ -1442,8 +1449,9 @@ gen_tics(struct axis *this, tic_callback callback)
 		    if (this->tictype == DT_TIMEDATE) {
 			mtic_user = time_tic_just(this->timelevel - 1, internal + mplace);
 			mtic_internal = mtic_user;
-		    } else if (nonlinear(this) && (def->type == TIC_COMPUTED)) {
-			/* FIXME: make up for bad calculation of ministart/ministep/miniend */
+		    } else if ((nonlinear(this) && (def->type == TIC_COMPUTED))
+			   ||  (nonlinear(this) && (def->type == TIC_SERIES && def->logscaling))) {
+			/* Make up for bad calculation of ministart/ministep/miniend */
 			double this_major = eval_link_function(this, internal);
 			double next_major = eval_link_function(this, internal+step);
 			mtic_user = this_major + mplace/miniend * (next_major - this_major);	
@@ -1661,7 +1669,7 @@ axis_output_tics(
 
 /* {{{ axis_set_scale_and_range() */
 void
-axis_set_scale_and_range(struct axis *axis, unsigned int lower, unsigned int upper)
+axis_set_scale_and_range(struct axis *axis, int lower, int upper)
 {
     axis->term_scale = (upper - lower) / (axis->max - axis->min);
     axis->term_lower = lower;
@@ -1759,7 +1767,7 @@ get_num_or_time(struct axis *axis)
 	struct tm tm;
 	double usec;
 	char *ss = try_to_get_string();
-	if (gstrptime(ss,timefmt,&tm,&usec))
+	if (gstrptime(ss, timefmt, &tm,&usec, &value) == DT_TIMEDATE)
 	    value = (double) gtimegm(&tm) + usec;
 	free(ss);
     } else {
@@ -2324,51 +2332,58 @@ int
 parse_range(AXIS_INDEX axis)
 {
     struct axis *this_axis = &axis_array[axis];
+    int dummy_token = -1;
 
-    if (equals(c_token, "[")) {
-	int dummy_token = -1;
-	c_token++;
-	/* If the range starts with "[var=" return the token of the named variable. */
-	if (isletter(c_token) && equals(c_token + 1, "=")) {
-		dummy_token = c_token;
-		c_token += 2;
-	}
-	this_axis->autoscale =
-		load_range(this_axis, &this_axis->min, &this_axis->max, this_axis->autoscale);
-#ifdef NONLINEAR_AXES
-	if (this_axis->linked_to_primary) {
-	    AXIS *primary = this_axis->linked_to_primary;
-	    primary->min = eval_link_function(primary, this_axis->min);
-	    primary->max = eval_link_function(primary, this_axis->max);
-	}
-#endif
-
-	/* DEBUG 2 Sep 2017
-	 * This handles (imperfectly) the problem case
-	 *    set link x2 via f(x) inv g(x)
-	 *    plot [x=min:max][] something that involves x2
-	 * Other cases of in-line range changes on a linked axis may fail
-	 */
-	else if (this_axis->linked_to_secondary) {
-	    AXIS *secondary = this_axis->linked_to_secondary;
-	    if (secondary->link_udf != NULL && secondary->link_udf->at != NULL)
-		clone_linked_axes(this_axis, secondary);
-	}
-
-	if (axis == SAMPLE_AXIS || axis == T_AXIS || axis == U_AXIS || axis == V_AXIS) {
-	    this_axis->SAMPLE_INTERVAL = 0;
-	    if (equals(c_token, ":")) {
-		c_token++;
-		this_axis->SAMPLE_INTERVAL = real_expression();
-	    }
-	}
-
-	if (!equals(c_token, "]"))
-	    int_error(c_token, "']' expected");
-	c_token++;
-	return dummy_token;
-    } else
+    /* No range present */
+    if (!equals(c_token, "["))
 	return 0;
+
+    /* Empty brackets serve as a place holder */
+    if (equals(c_token, "[]")) {
+	c_token += 2;
+	return 0;
+    }
+
+    /* If the range starts with "[var=" return the token of the named variable. */
+    c_token++;
+    if (isletter(c_token) && equals(c_token + 1, "=")) {
+	dummy_token = c_token;
+	c_token += 2;
+    }
+
+    this_axis->autoscale = load_range(this_axis, &this_axis->min, &this_axis->max,
+					this_axis->autoscale);
+
+    /* Nonlinear axis - find the linear range equivalent */
+    if (this_axis->linked_to_primary) {
+	AXIS *primary = this_axis->linked_to_primary;
+	clone_linked_axes(this_axis, primary);
+    }
+
+    /* This handles (imperfectly) the problem case
+     *    set link x2 via f(x) inv g(x)
+     *    plot [x=min:max][] something that involves x2
+     * Other cases of in-line range changes on a linked axis may fail
+     */
+    else if (this_axis->linked_to_secondary) {
+	AXIS *secondary = this_axis->linked_to_secondary;
+	if (secondary->link_udf != NULL && secondary->link_udf->at != NULL)
+	    clone_linked_axes(this_axis, secondary);
+    }
+
+    if (axis == SAMPLE_AXIS || axis == T_AXIS || axis == U_AXIS || axis == V_AXIS) {
+	this_axis->SAMPLE_INTERVAL = 0;
+	if (equals(c_token, ":")) {
+	    c_token++;
+	    this_axis->SAMPLE_INTERVAL = real_expression();
+	}
+    }
+
+    if (!equals(c_token, "]"))
+	int_error(c_token, "']' expected");
+    c_token++;
+
+    return dummy_token;
 }
 
 /* Called if an in-line range is encountered while inside a zoom command */
@@ -2388,7 +2403,7 @@ parse_skip_range()
 void
 clone_linked_axes(AXIS *axis1, AXIS *axis2)
 {
-    double testmin, testmax;
+    double testmin, testmax, scale;
     TBOOLEAN suspect = FALSE;
 
     /* DEBUG: This sanity check is only here for debugging */
@@ -2409,18 +2424,30 @@ clone_linked_axes(AXIS *axis1, AXIS *axis2)
 	axis2->min = eval_link_function(axis2, axis1->min);
 	axis2->max = eval_link_function(axis2, axis1->max);
 
-    /* Confirm that the inverse mapping actually works, at least at the endpoints */
+	/* Confirm that the inverse mapping actually works, at least at the endpoints.
+	 * We makes sure that inverse_f( f(x) ) = x at the edges of our plot
+	 * bounds, and if not, we throw a warning, and we try to be robust to
+	 * numerical-precision errors causing false-positive warnings. We look at
+	 * the error relative to a scaling:
+	 *      (inverse_f( f(x) ) - x) / scale
+	 * where the scale is the mean of (x(min edge of plot), x(max edge of
+	 * plot)). I.e. we only care about errors that are large on the scale of
+	 * the plot bounds we're looking at.
+	 */
+
 	if (isnan(axis2->set_min) || isnan(axis2->set_max))
 	    suspect = TRUE;
 	testmin = eval_link_function(axis1, axis2->set_min);
 	testmax = eval_link_function(axis1, axis2->set_max);
+	scale = (fabs(axis1->set_min) + fabs(axis1->set_max))/2.0;
+
 	if (isnan(testmin) || isnan(testmax))
 	    suspect = TRUE;
 	if (fabs(testmin - axis1->set_min) != 0
-	&&  fabs((testmin - axis1->set_min) / testmin) > 1.e-6)
+	&&  fabs((testmin - axis1->set_min) / scale) > 1.e-6)
 	    suspect = TRUE;
 	if (fabs(testmax - axis1->set_max) != 0
-	&&  fabs((testmax - axis1->set_max) / testmax) > 1.e-6)
+	&&  fabs((testmax - axis1->set_max) / scale) > 1.e-6)
 	    suspect = TRUE;
 
 	if (suspect) {
@@ -2432,8 +2459,21 @@ clone_linked_axes(AXIS *axis1, AXIS *axis2)
 		goto inverse_function_sanity_check;
 	    }
 	    int_warn(NO_CARET, "could not confirm linked axis inverse mapping function");
-	    fprintf(stderr,"\tmin: %g inv(via(min)): %g", axis1->set_min, testmin);
-	    fprintf(stderr,"  max: %g inv(via(max)): %g\n", axis1->set_max, testmax);
+	    fprintf(stderr, "\t");
+	    fprintf(stderr, "axis1: (set_min,min,testmin) = (%g,%g,%g); ",
+		    axis1->set_min,
+		    axis1->min,
+		    testmin);
+	    fprintf(stderr, "axis1: (set_max,max,testmax) = (%g,%g,%g); ",
+		    axis1->set_max,
+		    axis1->max,
+		    testmax);
+	    fprintf(stderr, "axis2: (set_min,min) = (%g,%g); ",
+		    axis1->set_min,
+		    axis1->min);
+	    fprintf(stderr, "axis2: (set_max,max) = (%g,%g)\n",
+		    axis1->set_max,
+		    axis1->max);
 	}
 }
 
@@ -2523,15 +2563,16 @@ extend_primary_ticrange(AXIS *axis)
 
 /*
  * As data is read in or functions evaluated, the min/max values are tracked
- * for the visible axes but not for the primary (linear) axes.
- * This routine fills in the primary min/max from the linked secondary axis.
+ * for the secondary (visible) axes but not for the linked primary (linear) axis.
+ * This routine fills in the primary min/max from the secondary axis.
  */
 void
 update_primary_axis_range(struct axis *secondary)
 {
     struct axis *primary = secondary->linked_to_primary;
 
-    if (nonlinear(secondary)) {
+    if (primary != NULL) {
+	/* nonlinear axis (secondary is visible; primary is hidden) */
 	primary->min = eval_link_function(primary, secondary->min);
 	primary->max = eval_link_function(primary, secondary->max);
 	primary->data_min = eval_link_function(primary, secondary->data_min);
@@ -2541,6 +2582,53 @@ update_primary_axis_range(struct axis *secondary)
 
 #endif
 
+/*
+ * Same thing but in the opposite direction.  We read in data on the primary axis
+ * and want the autoscaling on a linked secondary axis to match.
+ */
+void
+update_secondary_axis_range(struct axis *primary)
+{
+    struct axis *secondary = primary->linked_to_secondary;
+
+    if (secondary != NULL) {
+       secondary->min = eval_link_function(secondary, primary->min);
+       secondary->max = eval_link_function(secondary, primary->max);
+       secondary->data_min = eval_link_function(secondary, primary->data_min);
+       secondary->data_max = eval_link_function(secondary, primary->data_max);
+    }
+}
+
+/*
+ * gnuplot version 5.0 always maintained autoscaled range on x1
+ * specifically, transforming from x2 coordinates if necessary.
+ * In version 5.2 we track the x1 and x2 axis data limits separately.
+ * However if x1 and x2 are linked to each other we must reconcile
+ * their data limits before plotting.
+ */
+void
+reconcile_linked_axes(AXIS *primary, AXIS *secondary)
+{
+    double dummy;
+    int inrange = INRANGE;
+    if ((primary->autoscale & AUTOSCALE_BOTH) != AUTOSCALE_NONE
+    &&  primary->linked_to_secondary) {
+	double min_2_into_1 = eval_link_function(primary, secondary->data_min);
+	double max_2_into_1 = eval_link_function(primary, secondary->data_max);
+
+	/* Merge secondary min/max into primary data range */
+	ACTUAL_STORE_WITH_LOG_AND_UPDATE_RANGE(dummy, min_2_into_1, inrange, primary,
+		FALSE, NOOP, NOOP);
+	ACTUAL_STORE_WITH_LOG_AND_UPDATE_RANGE(dummy, max_2_into_1, inrange, primary,
+		FALSE, NOOP, NOOP);
+	(void)dummy;	/* Otherwise compilar complains about unused variable */
+
+	/* Take the result back the other way to update secondary */
+	secondary->min = eval_link_function(secondary, primary->min);
+	secondary->max = eval_link_function(secondary, primary->max);
+    }
+}
+
 
 /*
  * Check for linked-axis coordinate transformation given by command
